# Copyright (C) 2022 Canonical Ltd.
# Author: Nathan Pratta Teodosio
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Integration tests for the apport.crashdb_impl.github module."""

import unittest
from unittest.mock import ANY, Mock, patch

try:
    import apport.crashdb_impl.github

    IMPORT_ERROR = None
except ImportError as error:
    IMPORT_ERROR = error


@unittest.skipIf(IMPORT_ERROR, f"module not available: {IMPORT_ERROR}")
class TestGitHub(unittest.TestCase):
    # pylint: disable=missing-class-docstring,missing-function-docstring
    # pylint: disable=protected-access

    def setUp(self):
        self.crashdb = self._get_gh_database("Lorem", "Ipsum")
        self.crashdb_barren = self._get_gh_database(None, None)

        self.message_cb = Mock()
        self.github = apport.crashdb_impl.github.Github(
            self.crashdb.app_id, self.message_cb
        )

        self.api_auth_return_value = {
            "verification_uri": None,
            "user_code": 123,
            "device_code": "d123",
            "interval": 1,
            "expires_in": 20,
        }

    def test__format_report(self):
        data = {"bold1": "normal1", "bold2": "normal2"}
        expected_body = "**bold1**\nnormal1\n\n**bold2**\nnormal2\n\n"

        result = self.crashdb._format_report(data)
        self.assertTrue("body" in result and result["body"] == expected_body)
        self.assertTrue("title" in result)
        self.assertTrue(self.crashdb.labels == set(result["labels"]))

    @patch("apport.crashdb_impl.github.Github.api_authentication")
    @patch("apport.crashdb_impl.github.Github.api_open_issue")
    @patch("apport.crashdb_impl.github.Github.authentication_complete")
    def test_upload(self, mock_auth, mock_api, mock_api_auth):
        mock_api.return_value = {"html_url": "doesntmatterhere"}
        mock_auth.return_value = True
        mock_api_auth.return_value = self.api_auth_return_value
        nodata = {}
        snapdata = {"SnapGitOwner": "gimli", "SnapGitName": "axe"}

        with self.github as github:
            self.crashdb.github = github
            self.crashdb_barren.github = github

            # Snap fields are not set and the database specifies no
            # norepository_{owner,name}
            self.assertRaises(
                RuntimeError, self.crashdb_barren.upload, nodata, None, self.message_cb
            )

            # Snap fields are not set and the database specifies
            # norepository_{owner,name}
            self.crashdb.upload(nodata, None, self.message_cb)
            mock_api.assert_called_with(
                self.crashdb.repository_owner, self.crashdb.repository_name, ANY
            )

            # Snap fields are set and the database specifies
            # norepository_{owner,name}
            self.crashdb_barren.upload(snapdata, None, self.message_cb)
            mock_api.assert_called_with(
                snapdata["SnapGitOwner"], snapdata["SnapGitName"], ANY
            )

            # Snap fields are set and the database specifies
            # repository_{owner,name}: The database specs takes precedence.
            self.crashdb.upload(snapdata, None, self.message_cb)
            mock_api.assert_called_with(
                self.crashdb.repository_owner, self.crashdb.repository_name, ANY
            )

    @patch("apport.crashdb_impl.github.Github.api_authentication")
    def test_authentication_complete(self, mock_api):
        base_response = self.api_auth_return_value
        mocked = [base_response.copy() for i in range(7)]
        mocked[2]["error"] = "foo"
        mocked[3]["error"] = "authorization_pending"
        mocked[4]["error"] = "slow_down"
        mocked[5]["access_token"] = "token"
        mocked[6]["expires_in"] = -100
        mock_api.side_effect = mocked

        # No __enter__, no authentication data
        self.assertRaises(RuntimeError, self.github.authentication_complete)
        self.message_cb.assert_not_called()
        with self.github as github:
            # Error message missing
            self.assertRaises(RuntimeError, github.authentication_complete)
            self.message_cb.assert_called_with("Login required", ANY)
            # Error message awry
            self.assertRaises(RuntimeError, github.authentication_complete)
            # Still not authorized
            self.assertFalse(github.authentication_complete())
            # Slow down!
            self.assertFalse(github.authentication_complete())
            # Access token OK
            self.assertTrue(github.authentication_complete())
        with self.github as github:
            # Expired (requires __enter__ again to update __expiry).
            self.assertRaises(RuntimeError, github.authentication_complete)
            self.message_cb.assert_called_with(
                "Failed login", "Github authentication expired. Please try again."
            )

    def test_not_implemented_methods(self):
        ni = NotImplementedError
        self.assertRaises(ni, self.crashdb._mark_dup_checked, None, None)
        self.assertRaises(ni, self.crashdb.can_update, None)
        self.assertRaises(ni, self.crashdb.close_duplicate, None, None, None)
        self.assertRaises(ni, self.crashdb.download, None)
        self.assertRaises(ni, self.crashdb.duplicate_of, None)
        self.assertRaises(ni, self.crashdb.get_affected_packages, None)
        self.assertRaises(ni, self.crashdb.get_distro_release, None)
        self.assertRaises(ni, self.crashdb.get_dup_unchecked)
        self.assertRaises(ni, self.crashdb.get_fixed_version, None)
        self.assertRaises(ni, self.crashdb.get_id_url, None, None)
        self.assertRaises(ni, self.crashdb.get_unfixed)
        self.assertRaises(ni, self.crashdb.get_unretraced)
        self.assertRaises(ni, self.crashdb.is_reporter, None)
        self.assertRaises(ni, self.crashdb.mark_regression, None, None)
        self.assertRaises(ni, self.crashdb.mark_retrace_failed, None, None)
        self.assertRaises(ni, self.crashdb.mark_retraced, None)
        self.assertRaises(ni, self.crashdb.update, None, None, None, None)

    @staticmethod
    def _get_gh_database(repository_owner, repository_name):
        return apport.crashdb_impl.github.CrashDatabase(
            None,
            {
                "impl": "github",
                "repository_owner": repository_owner,
                "repository_name": repository_name,
                "github_app_id": "a654870577ad2a2ab5b1",
                "labels": ["apport"],
            },
        )



#!/usr/bin/python3

"""Dump ACPI tables."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import os
import stat
import sys


def dump_acpi_table(filename, tablename, out):
    """Dump a single ACPI table."""
    if not os.access(filename, os.R_OK):
        return

    out.write(f"{tablename[0:4]} @ 0x0000000000000000\n")
    n = 0
    with open(filename, "rb") as f:
        hex_str = ""
        try:
            byte = f.read(1)
            while byte != b"":
                val = ord(byte)
                if (n & 15) == 0:
                    if n > 65535:
                        hex_str = f"   {n:04X}: "
                    else:
                        hex_str = f"    {n:04X}: "
                    ascii_str = ""

                hex_str = f"{hex_str}{val:02X} "

                if (val < 32) or (val > 126):
                    ascii_str = f"{ascii_str}."
                else:
                    ascii_str = ascii_str + chr(val)
                n = n + 1
                if (n & 15) == 0:
                    out.write(f"{hex_str} {ascii_str}\n")
                byte = f.read(1)
        finally:
            if (n % 16) != 0:
                hex_str += "   " * (16 - n % 16)
                out.write(f"{hex_str} {ascii_str}\n")

    out.write("\n")


def dump_acpi_tables(path, out):
    """Dump ACPI tables."""
    tables = os.listdir(path)
    for tablename in tables:
        pathname = os.path.join(path, tablename)
        mode = os.stat(pathname).st_mode
        if stat.S_ISDIR(mode):
            dump_acpi_tables(pathname, out)
        else:
            dump_acpi_table(pathname, tablename, out)


if os.path.isdir("/sys/firmware/acpi/tables"):
    dump_acpi_tables("/sys/firmware/acpi/tables", sys.stdout)



"""Unit tests for the apport.fileutils module."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name,missing-class-docstring,missing-function-docstring

import io
import os
import pathlib
import tempfile
import time
import unittest
import unittest.mock
from unittest.mock import MagicMock

import apport.fileutils
import apport.packaging


class T(unittest.TestCase):
    @unittest.mock.patch("apport.fileutils.packaging.get_files")
    def test_find_package_desktopfile_deleted(self, get_files_mock):
        """find_package_desktopfile() for a deleted desktop file."""
        get_files_mock.return_value = [
            "/usr/share/applications/non-existing/blueman.desktop"
        ]
        self.assertIsNone(apport.fileutils.find_package_desktopfile("blueman"))
        get_files_mock.assert_called_once_with("blueman")

    def test_likely_packaged(self):
        """likely_packaged()"""
        self.assertEqual(apport.fileutils.likely_packaged("/bin/bash"), True)
        self.assertEqual(apport.fileutils.likely_packaged("/usr/bin/foo"), True)
        self.assertEqual(apport.fileutils.likely_packaged("/usr/local/bin/foo"), False)
        self.assertEqual(apport.fileutils.likely_packaged("/home/test/bin/foo"), False)
        self.assertEqual(apport.fileutils.likely_packaged("/tmp/foo"), False)
        # ignore crashes in /var/lib (LP#122859, LP#414368)
        self.assertEqual(apport.fileutils.likely_packaged("/var/lib/foo"), False)

    @unittest.mock.patch("apport.fileutils.get_boot_id")
    def test_get_core_path_missing_timestamp(self, get_boot_id_mock: MagicMock) -> None:
        """Test get_core_path() with missing timestamp parameter."""
        get_boot_id_mock.return_value = "boot-id"
        pid = os.getpid()
        core_name, core_path = apport.fileutils.get_core_path(pid, "/usr/bin/python3")

        with open(f"/proc/{pid}/stat", encoding="utf-8") as stat_file:
            contents = stat_file.read()
        starttime = apport.fileutils.get_starttime(contents)
        expected = f"core._usr_bin_python3.{os.getuid()}.boot-id.{pid}.{starttime}"
        expected_path = os.path.join(apport.fileutils.core_dir, expected)
        self.assertEqual(core_name, expected)
        self.assertEqual(core_path, expected_path)
        get_boot_id_mock.assert_called_once_with()

    @unittest.mock.patch("apport.fileutils.get_boot_id")
    def test_get_core_path_proc_pid_fd(self, get_boot_id_mock: MagicMock) -> None:
        """Test get_core_path() with proc_pid_fd provided."""
        get_boot_id_mock.return_value = "boot-id"
        fake_pid = 67890
        now = int(time.clock_gettime(time.CLOCK_BOOTTIME) * 100)
        with tempfile.TemporaryDirectory() as tmpdir:
            stat = pathlib.Path(tmpdir) / "stat"
            stat.write_text(f"{fake_pid} (name) ?{' x' * 18} {now} ...\n")
            exe = pathlib.Path(tmpdir) / "exe"
            exe.symlink_to("/usr/bin/myprog")
            fd = os.open(tmpdir, os.O_RDONLY | os.O_PATH | os.O_DIRECTORY)
            try:
                (core_name, core_path) = apport.fileutils.get_core_path(
                    fake_pid, uid=42, proc_pid_fd=fd
                )
            finally:
                os.close(fd)

        expected = f"core._usr_bin_myprog.42.boot-id.{fake_pid}.{now}"
        expected_path = os.path.join(apport.fileutils.core_dir, expected)
        self.assertEqual(core_name, expected)
        self.assertEqual(core_path, expected_path)
        get_boot_id_mock.assert_called_once_with()

    def test_get_login_defs(self) -> None:
        """Test get_login_defs()."""
        apport.fileutils.get_login_defs.cache_clear()
        open_mock = unittest.mock.mock_open(
            read_data="#\n"
            "# /etc/login.defs\n"
            "#\n"
            "MAIL_DIR        /var/mail\n"
            "#MAIL_FILE      .mail\n"
            "UID_MIN			 1000\n"
            "UID_MAX			60000\n"
        )
        with unittest.mock.patch("builtins.open", open_mock):
            login_defs = apport.fileutils.get_login_defs()
        open_mock.assert_called_once()
        self.assertEqual(
            login_defs, {"MAIL_DIR": "/var/mail", "UID_MIN": "1000", "UID_MAX": "60000"}
        )

    def test_get_login_defs_missing(self) -> None:
        """Test get_login_defs() with /etc/login.defs not present."""
        apport.fileutils.get_login_defs.cache_clear()
        open_mock = MagicMock(side_effect=FileNotFoundError)
        with unittest.mock.patch("builtins.open", open_mock):
            login_defs = apport.fileutils.get_login_defs()
        open_mock.assert_called_once()
        self.assertEqual(login_defs, {})

    def test_get_sys_gid_max(self) -> None:
        """Test get_sys_gid_max()."""
        apport.fileutils.get_login_defs.cache_clear()
        open_mock = unittest.mock.mock_open(read_data="SYS_GID_MAX 1337\n")
        with unittest.mock.patch("builtins.open", open_mock):
            sys_gid_max = apport.fileutils.get_sys_gid_max()
        open_mock.assert_called_once()
        self.assertEqual(sys_gid_max, 1337)

    def test_get_sys_gid_max_default(self) -> None:
        """Test get_sys_gid_max() returning the default."""
        apport.fileutils.get_login_defs.cache_clear()
        open_mock = MagicMock(side_effect=FileNotFoundError)
        with unittest.mock.patch("builtins.open", open_mock):
            sys_gid_max = apport.fileutils.get_sys_gid_max()
        open_mock.assert_called_once()
        self.assertEqual(sys_gid_max, 999)

    def test_get_sys_uid_max(self) -> None:
        """Test get_sys_uid_max()."""
        apport.fileutils.get_login_defs.cache_clear()
        open_mock = unittest.mock.mock_open(read_data="SYS_UID_MAX 937\n")
        with unittest.mock.patch("builtins.open", open_mock):
            sys_gid_max = apport.fileutils.get_sys_uid_max()
        open_mock.assert_called_once()
        self.assertEqual(sys_gid_max, 937)

    def test_get_sys_uid_max_default(self) -> None:
        """Test get_sys_uid_max() returning the default."""
        apport.fileutils.get_login_defs.cache_clear()
        open_mock = unittest.mock.mock_open(read_data="MAIL_DIR        /var/mail\n")
        with unittest.mock.patch("builtins.open", open_mock):
            sys_gid_max = apport.fileutils.get_sys_uid_max()
        open_mock.assert_called_once()
        self.assertEqual(sys_gid_max, 999)

    def test_get_process_environ(self) -> None:
        open_mock = unittest.mock.mock_open(
            read_data="SHELL=/bin/bash\0TERM=xterm-256color\0LANGUAGE=en\0"
        )
        with unittest.mock.patch("builtins.open", open_mock):
            environ = apport.fileutils.get_process_environ(1337)
        open_mock.assert_called_once()
        self.assertEqual(
            environ, {"LANGUAGE": "en", "SHELL": "/bin/bash", "TERM": "xterm-256color"}
        )

    def test_get_process_environ_empty(self) -> None:
        open_mock = unittest.mock.mock_open(read_data="")
        with unittest.mock.patch("builtins.open", open_mock):
            environ = apport.fileutils.get_process_environ(1337)
        open_mock.assert_called_once()
        self.assertEqual(environ, {})

    def test_get_process_environ_malformed(self) -> None:
        """Test get_process_environ() with malformed /proc/<pid>/environ.

        On Ubuntu 22.10 the mattermost snap had a malformed environment
        file (see mocked value in this test). /proc/<pid>/cmdline contained:
        /snap/mattermost-desktop/578/opt/Mattermost/mattermost-desktop
        --no-sandbox --disable-seccomp-filter-sandbox --enable-crashpad
        """
        open_mock = unittest.mock.mock_open(read_data="--enable-crashpad" + "\0" * 6355)
        with unittest.mock.patch("builtins.open", open_mock):
            environ = apport.fileutils.get_process_environ(1337)
        open_mock.assert_called_once()
        self.assertEqual(environ, {})

    def test_get_recent_crashes(self):
        """get_recent_crashes()"""
        # incomplete fields
        r = io.BytesIO(b"""ProblemType: Crash""")
        self.assertEqual(apport.fileutils.get_recent_crashes(r), 0)

        r = io.BytesIO(b"""ProblemType: Crash\nDate: Wed Aug 01 00:00:01 1990\n""")
        self.assertEqual(apport.fileutils.get_recent_crashes(r), 0)

        # ancient report
        r = io.BytesIO(
            b"ProblemType: Crash\n"
            b"Date: Wed Aug 01 00:00:01 1990\n"
            b"CrashCounter: 3\n"
        )
        self.assertEqual(apport.fileutils.get_recent_crashes(r), 0)

        # old report (one day + one hour ago)
        date = time.ctime(time.mktime(time.localtime()) - 25 * 3600)
        r = io.BytesIO(f"ProblemType: Crash\nDate: {date}\nCrashCounter: 3\n".encode())
        self.assertEqual(apport.fileutils.get_recent_crashes(r), 0)

        # current report (one hour ago)
        date = time.ctime(time.mktime(time.localtime()) - 3600)
        r = io.BytesIO(f"ProblemType: Crash\nDate: {date}\nCrashCounter: 3\n".encode())
        self.assertEqual(apport.fileutils.get_recent_crashes(r), 3)

    def test_get_dbus_socket(self):
        """get_dbus_socket()"""
        tests = [
            ("unix:path=/run/user/1000/bus", "/run/user/1000/bus"),
            ("unix:path=/run/user/1000/bus;unix:path=/run/user/0/bus", None),
            ("unix:path=%2Frun/user/1000/bus", None),
            ("unix:path=/run/user/1000/bus,path=/run/user/0/bus", None),
            ("unix:path=/etc/passwd", None),
            ("unix:path=/run/user/../../etc/passwd", None),
            ("unix:path=/run/user/1000/bus=", None),
            ("", None),
            ("tcp:host=localhost,port=8100", None),
        ]

        for addr, result in tests:
            self.assertEqual(apport.fileutils.get_dbus_socket(addr), result)

    def test_get_starttime(self):
        """get_starttime()"""
        template = (
            "2022799 (%s) S 1834041 2022799 2022799 34820 "
            "2022806 4194304 692 479 0 0 2 0 0 0 20 0 1 0 "
            "34034895 13750272 1270 18446744073709551615 "
            "94876723015680 94876723738373 140731174338544 "
            "0 0 0 65536 3670020 1266777851 1 0 0 17 7 0 0 0 0 0 "
            "94876723969264 94876724016644 94876748173312 "
            "140731174345133 140731174345138 140731174345138 "
            "140731174346734 0\n"
        )

        for name in (
            "goodname",
            "bad name",
            "(badname",
            "badname)",
            "(badname)",
            "((badname))",
            "badname(",
            ")badname",
            "bad) name",
            "bad) name\\",
        ):
            starttime = apport.fileutils.get_starttime(template % name)
            self.assertEqual(starttime, 34034895)

    def test_get_uid_and_gid(self):
        """get_uid_and_gid()"""
        # Python 3's open uses universal newlines, which means all
        # line endings get replaced with \n, so we don't need to test
        # different line ending combinations
        template = (
            "Name:\t%s\n"
            "Umask:\t0002\n"
            "State:\tS (sleeping)\n"
            "Tgid:\t2288405\n"
            "Ngid:\t0\n"
            "Pid:\t2288405\n"
            "PPid:\t2167353\n"
            "TracerPid:\t0\n"
            "Uid:\t1000\t1000\t1000\t1000\n"
            "Gid:\t2000\t2000\t2000\t2000\n"
            "FDSize:\t256\n"
            "Groups:\t4 20 24 27 30 46 108 120 131 132 135 137 1000 \n"
            "NStgid:\t2288405\n"
            "NSpid:\t2288405\n"
            "NSpgid:\t2288405\n"
            "NSsid:\t2288405\n"
            "VmPeak:\t   13428 kB\n"
            "VmSize:\t   13428 kB\n"
            "VmLck:\t       0 kB\n"
            "VmPin:\t       0 kB\n"
            "VmHWM:\t    5200 kB\n"
            "VmRSS:\t    5200 kB\n"
            "RssAnon:\t    1700 kB\n"
            "RssFile:\t    3500 kB\n"
            "RssShmem:\t       0 kB\n"
            "VmData:\t    1600 kB\n"
            "VmStk:\t     132 kB\n"
            "VmExe:\t     708 kB\n"
            "VmLib:\t    1748 kB\n"
            "VmPTE:\t      68 kB\n"
            "VmSwap:\t       0 kB\n"
            "HugetlbPages:\t       0 kB\n"
            "CoreDumping:\t0\n"
            "THP_enabled:\t1\n"
            "Threads:\t1\n"
            "SigQ:\t0/62442\n"
            "SigPnd:\t0000000000000000\n"
            "ShdPnd:\t0000000000000000\n"
            "SigBlk:\t0000000000010000\n"
            "SigIgn:\t0000000000380004\n"
            "SigCgt:\t000000004b817efb\n"
            "CapInh:\t0000000000000000\n"
            "CapPrm:\t0000000000000000\n"
            "CapEff:\t0000000000000000\n"
            "CapBnd:\t000000ffffffffff\n"
            "CapAmb:\t0000000000000000\n"
            "NoNewPrivs:\t0\n"
            "Seccomp:\t0\n"
            "Speculation_Store_Bypass:\tthread vulnerable\n"
            "Cpus_allowed:\tff\n"
            "Cpus_allowed_list:\t0-7\n"
            "Mems_allowed:\t00000000,00000000,00000000,00000000,"
            "00000000,00000000,00000000,00000000,00000000,"
            "00000000,00000000,00000000,00000000,00000000,"
            "00000000,00000000,00000000,00000000,00000000,"
            "00000000,00000000,00000000,00000000,00000000,"
            "00000000,00000000,00000000,00000000,00000000,"
            "00000000,00000000,00000001\n"
            "Mems_allowed_list:\t0\n"
            "voluntary_ctxt_switches:\t133\n"
            "nonvoluntary_ctxt_switches:\t0\n"
        )

        for name in (
            "goodname",
            "Uid:",
            "\nUid:",
            "Uid: 1000 1000 1000\n",
            "a\nUid: 0\nGid: 0",
            "a\nUid: 0\nGid:",
        ):
            (uid, gid) = apport.fileutils.get_uid_and_gid(template % name)
            self.assertEqual(uid, 1000)
            self.assertEqual(gid, 2000)



# Copyright (C) 2022 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Add image build information if available."""

import os

_BUILD_INFO_KEY_MAPPING = {"build_name": "CloudBuildName", "serial": "CloudSerial"}


def _add_cloud_build_info(report):
    """Add cloud build information if available.

    /etc/cloud/build.info is generated by livecd-rootfs.
    """
    if not os.path.isfile("/etc/cloud/build.info"):
        return
    report.add_tags(["cloud-image"])
    with open("/etc/cloud/build.info", encoding="utf-8") as build_info:
        for line in build_info:
            try:
                key, value = line.split(":", 1)
            except ValueError:
                continue
            if key not in _BUILD_INFO_KEY_MAPPING:
                continue
            report[_BUILD_INFO_KEY_MAPPING[key]] = value.strip()


def add_info(report, unused_ui):
    """Add image build information if available."""
    _add_cloud_build_info(report)



"""System tests for data/apport."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import errno
import os
import resource
import shutil
import signal
import subprocess
import sys
import tempfile
import time
import unittest
from unittest.mock import MagicMock, patch

import psutil

import apport.fileutils
from tests.helper import (
    get_init_system,
    pidof,
    skip_if_command_is_missing,
    wait_for_process_to_appear,
    wait_for_sleeping_state,
)
from tests.paths import (
    get_data_directory,
    is_local_source_directory,
    local_test_environment,
)


class T(unittest.TestCase):
    # pylint: disable=missing-class-docstring,missing-function-docstring
    # pylint: disable=protected-access
    TEST_EXECUTABLE = os.path.realpath("/bin/sleep")
    TEST_ARGS = ["86400"]

    @classmethod
    def setUpClass(cls):
        cls.orig_environ = os.environ.copy()
        os.environ |= local_test_environment()

        with open("/proc/sys/kernel/core_pattern", encoding="utf-8") as f:
            core_pattern = f.read().strip()
        if core_pattern[0] == "|":
            cls.apport_path = core_pattern[1:].split()[0]
        else:
            cls.apport_path = None
        if is_local_source_directory():
            cls.apport_path = get_data_directory() / "apport"

        cls.all_reports = apport.fileutils.get_all_reports()

        # ensure we don't inherit an ignored SIGQUIT
        signal.signal(signal.SIGQUIT, signal.SIG_DFL)

        orig_home = os.getenv("HOME")
        if orig_home is not None:
            del os.environ["HOME"]
        cls.ifpath = os.path.expanduser(apport.report._ignore_file)
        if orig_home is not None:
            os.environ["HOME"] = orig_home

        cls.orig_cwd = os.getcwd()
        cls.orig_core_dir = apport.fileutils.core_dir
        cls.orig_ignore_file = apport.report._ignore_file
        cls.orig_report_dir = apport.fileutils.report_dir

    @classmethod
    def tearDownClass(cls):
        os.environ.clear()
        os.environ.update(cls.orig_environ)
        apport.fileutils.core_dir = cls.orig_core_dir
        apport.report._ignore_file = cls.orig_ignore_file
        apport.fileutils.report_dir = cls.orig_report_dir
        os.chdir(cls.orig_cwd)

    def setUp(self):
        if self.apport_path is None:
            self.skipTest(
                "kernel crash dump helper is not active; please enable"
                " before running this test"
            )

        if self.all_reports:
            self.skipTest(
                "Please remove all crash reports from /var/crash/ for this"
                " test suite:\n  %s\n" % "\n  ".join(self.all_reports)
            )

        # use local report dir
        self.report_dir = tempfile.mkdtemp()
        os.environ["APPORT_REPORT_DIR"] = self.report_dir
        apport.fileutils.report_dir = self.report_dir

        self.workdir = tempfile.mkdtemp()

        apport.fileutils.core_dir = os.path.join(self.workdir, "coredump")
        os.mkdir(apport.fileutils.core_dir)
        os.environ["APPORT_COREDUMP_DIR"] = apport.fileutils.core_dir

        os.environ["APPORT_IGNORE_FILE"] = os.path.join(
            self.workdir, "apport-ignore.xml"
        )
        apport.report._ignore_file = os.environ["APPORT_IGNORE_FILE"]

        os.environ["APPORT_LOCK_FILE"] = os.path.join(self.workdir, "apport.lock")

        # move aside current ignore file
        if os.path.exists(self.ifpath):
            os.rename(self.ifpath, f"{self.ifpath}.apporttest")

        # do not write core files by default
        resource.setrlimit(resource.RLIMIT_CORE, (0, -1))

        # that's the place where to put core dumps, etc.
        os.chdir("/tmp")

        # expected report name for test executable report
        self.test_report = os.path.join(
            apport.fileutils.report_dir,
            f"{self.TEST_EXECUTABLE.replace('/', '_')}.{os.getuid()}.crash",
        )

        self.running_test_executables = pidof(self.TEST_EXECUTABLE)

    def tearDown(self):
        shutil.rmtree(self.report_dir)
        shutil.rmtree(self.workdir)

        # clean up our ignore file
        if os.path.exists(self.ifpath):
            os.unlink(self.ifpath)
        orig_ignore_file = f"{self.ifpath}.apporttest"
        if os.path.exists(orig_ignore_file):
            os.rename(orig_ignore_file, self.ifpath)

        # permit tests to leave behind test_report, but nothing else
        if os.path.exists(self.test_report):
            apport.fileutils.delete_report(self.test_report)
        unexpected_reports = apport.fileutils.get_all_reports()
        for r in unexpected_reports:
            apport.fileutils.delete_report(r)
        self.assertEqual(unexpected_reports, [])

    def test_limit_size(self):
        """Core dumps are capped on available memory size."""
        # determine how much data we have to pump into apport in order to make
        # sure that it will refuse the core dump
        r = apport.Report()
        with open("/proc/meminfo", "rb") as f:
            r.load(f)
        totalmb = int(r["MemFree"].split()[0]) + int(r["Cached"].split()[0])
        totalmb = int(totalmb / 1024)
        r = None

        test_proc = self.create_test_process()
        try:
            with subprocess.Popen(
                [
                    self.apport_path,
                    "-p",
                    str(test_proc.pid),
                    "-s",
                    "42",
                    "-c",
                    "0",
                    "-d",
                    "1",
                ],
                stdin=subprocess.PIPE,
                stderr=subprocess.PIPE,
            ) as app:
                # pipe an entire total memory size worth of spaces into it,
                # which must be bigger than the 'usable' memory size. apport
                # should digest that and the report should not have a core
                # dump; NB that this should error out with a SIGPIPE when
                # apport aborts reading from stdin
                onemb = b" " * 1048576
                while totalmb > 0:
                    if totalmb & 255 == 0:
                        # Print a dot every 256 MiB
                        sys.stderr.write(".")
                        sys.stderr.flush()
                    try:
                        app.stdin.write(onemb)
                    except OSError as error:
                        if error.errno == errno.EPIPE:
                            break
                        raise
                    totalmb -= 1
                err = app.communicate()[1]
            self.assertEqual(app.returncode, 0, err)
            onemb = None
        finally:
            test_proc.kill()
            test_proc.wait()

        reports = apport.fileutils.get_all_reports()
        self.assertEqual(reports, [self.test_report])

        pr = apport.Report()
        with open(reports[0], "rb") as f:
            pr.load(f)
        os.unlink(reports[0])

        self.assertEqual(pr["Signal"], "42")
        self.assertEqual(pr["ExecutablePath"], self.TEST_EXECUTABLE)
        self.assertNotIn("CoreDump", pr)
        # FIXME: sometimes this is empty!?
        if err:
            self.assertRegex(
                err.decode(),
                f"core dump exceeded.*dropped from .*"
                f"{os.path.basename(self.TEST_EXECUTABLE)}\\..*\\.crash",
            )

    @unittest.skipIf(
        get_init_system() != "systemd", "running init system is not systemd"
    )
    @skip_if_command_is_missing("systemd-run")
    @unittest.skipIf(os.geteuid() != 0, "this test needs to be run as root")
    def test_crash_system_slice(self):
        """Report generation for a protected process running in the system
        slice"""
        apport.fileutils.report_dir = self.orig_report_dir
        self.test_report = os.path.join(
            apport.fileutils.report_dir,
            f"{self.TEST_EXECUTABLE.replace('/', '_')}.{os.getuid()}.crash",
        )

        system_run = self.create_test_process(
            command="/usr/bin/systemd-run",
            args=[
                "-t",
                "-q",
                "--slice=system.slice",
                "-p",
                "ProtectSystem=true",
                self.TEST_EXECUTABLE,
            ]
            + self.TEST_ARGS,
        )
        try:
            sleep_pid = wait_for_process_to_appear(
                self.TEST_EXECUTABLE, self.running_test_executables
            )
            wait_for_sleeping_state(sleep_pid)
            os.kill(sleep_pid, signal.SIGSEGV)

            self.wait_for_no_instance_running(self.TEST_EXECUTABLE)
            self.wait_for_apport_to_finish()
        finally:
            system_run.kill()
            system_run.wait()

        # check crash report
        with open("/var/log/apport.log", encoding="utf-8") as logfile:
            apport_log = logfile.read().strip()
        reports = apport.fileutils.get_all_reports()
        self.assertEqual(reports, [self.test_report], f"Apport log:\n{apport_log}")
        self.assertEqual(
            reports[0], f"/var/crash/{self.TEST_EXECUTABLE.replace('/', '_')}.0.crash"
        )

    def create_test_process(self, command=None, uid=None, args=None):
        """Spawn test executable.

        Wait until it is fully running, and return its process.
        """
        if command is None:
            command = self.TEST_EXECUTABLE
        if args is None:
            args = self.TEST_ARGS

        assert os.access(command, os.X_OK), f"{command} is not executable"

        env = os.environ.copy()
        # set UTF-8 environment variable, to check proper parsing in apport
        os.putenv("utf8trap", b"\xc3\xa0\xc3\xa4")
        # caller needs to call .wait(), pylint: disable=consider-using-with
        # False positive, see https://github.com/PyCQA/pylint/issues/7092
        process = subprocess.Popen(  # pylint: disable=unexpected-keyword-arg
            [command] + args, env=env, stdout=subprocess.DEVNULL, user=uid
        )

        # wait until child process has execv()ed properly
        while True:
            with open(f"/proc/{process.pid}/cmdline", encoding="utf-8") as f:
                cmdline = f.read()
            if "test_signal" in cmdline:
                time.sleep(0.1)
            else:
                break

        # sleep command needs extra time to get into "sleeping" state
        if command == self.TEST_EXECUTABLE:
            wait_for_sleeping_state(process.pid)

        return process

    def test_create_test_sleep_process(self) -> None:
        """Test create_test_process() helper method with sleep process."""
        proc = self.create_test_process()
        try:
            self.assertEqual(psutil.Process(proc.pid).status(), "sleeping")
        finally:
            proc.kill()
            proc.wait()

    @patch("tests.helper.wait_for_sleeping_state")
    def test_create_test_non_sleep_process(self, wait_sleep_mock: MagicMock) -> None:
        """Test create_test_process() helper method with non-sleep process."""
        with (
            patch("os.access"),
            patch("subprocess.Popen"),
            patch("tests.system.test_signal_crashes.open"),
        ):
            self.create_test_process(command="echo")

        wait_sleep_mock.assert_not_called()

    def wait_for_apport_to_finish(self, timeout_sec=10.0):
        self.wait_for_no_instance_running(self.apport_path, timeout_sec)

    def wait_for_no_instance_running(self, program, timeout_sec=10.0):
        while timeout_sec > 0:
            if not pidof(program) - self.running_test_executables:
                break
            time.sleep(0.2)
            timeout_sec -= 0.2
        else:
            self.fail(f"Timeout exceeded, but {program} is still running.")



"""Unit tests for the apport.ui module."""

import os
import tempfile
import textwrap
import unittest
import unittest.mock
from gettext import gettext as _
from unittest.mock import MagicMock

import apport.ui


class TestUI(unittest.TestCase):
    """Unit tests for apport.ui."""

    crashdb_conf: str

    @classmethod
    def setUpClass(cls) -> None:
        # pylint: disable-next=consider-using-with
        crashdb_conf_file = tempfile.NamedTemporaryFile("w+")
        cls.addClassCleanup(crashdb_conf_file.close)
        crashdb_conf_file.write(
            textwrap.dedent(
                """\
            default = 'testsuite'
            databases = {
                'testsuite': {
                    'impl': 'memory',
                    'bug_pattern_url': None,
                },
                'debug': {
                    'impl': 'memory',
                    'distro': 'debug',
                },
            }
            """
            )
        )
        crashdb_conf_file.flush()
        cls.crashdb_conf = crashdb_conf_file.name

    def setUp(self) -> None:
        self.orig_environ = os.environ.copy()
        os.environ["APPORT_CRASHDB_CONF"] = self.crashdb_conf
        self.ui = apport.ui.UserInterface([])

        # Mock environment for run_as_real_user() to not being called via sudo/pkexec
        os.environ.pop("SUDO_UID", None)
        os.environ.pop("PKEXEC_UID", None)

    def tearDown(self) -> None:
        os.environ.clear()
        os.environ.update(self.orig_environ)

    @unittest.mock.patch("subprocess.run")
    @unittest.mock.patch("webbrowser.open")
    def test_open_url(self, open_mock: MagicMock, run_mock: MagicMock) -> None:
        """Test successful UserInterface.open_url() without pkexec/sudo."""
        self.ui.open_url("https://example.com")

        run_mock.assert_called_once_with(
            ["xdg-open", "https://example.com"], check=False
        )
        open_mock.assert_not_called()

    @unittest.mock.patch("subprocess.run")
    @unittest.mock.patch("webbrowser.open")
    def test_open_url_webbrowser_fallback(
        self, open_mock: MagicMock, run_mock: MagicMock
    ) -> None:
        """Test UserInterface.open_url() to fall back to webbrowser.open()."""
        run_mock.side_effect = FileNotFoundError
        open_mock.return_value = True

        self.ui.open_url("https://example.net")

        run_mock.assert_called_once_with(
            ["xdg-open", "https://example.net"], check=False
        )
        open_mock.assert_called_once_with("https://example.net", new=1, autoraise=True)

    @unittest.mock.patch("subprocess.run")
    @unittest.mock.patch("webbrowser.open")
    @unittest.mock.patch("apport.ui.UserInterface.ui_error_message")
    def test_open_url_webbrowser_fails(
        self, error_message_mock: MagicMock, open_mock: MagicMock, run_mock: MagicMock
    ) -> None:
        """Test UserInterface.open_url() with webbrowser.open() returning False."""
        run_mock.side_effect = FileNotFoundError
        open_mock.return_value = False

        self.ui.open_url("https://example.org")

        run_mock.assert_called_once_with(
            ["xdg-open", "https://example.org"], check=False
        )
        open_mock.assert_called_once_with("https://example.org", new=1, autoraise=True)
        error_message_mock.assert_called_once_with(
            _("Unable to start web browser"),
            _("Unable to start web browser to open %s.") % ("https://example.org"),
        )



"""Unit tests for the apport.hookutils module."""

import datetime
import os
import re
import subprocess
import unittest
import unittest.mock
from unittest.mock import MagicMock, Mock

import apport.hookutils


class TestHookutils(unittest.TestCase):
    # pylint: disable=missing-function-docstring
    """Test apport.hookutils module."""

    maxDiff = None

    @unittest.mock.patch("os.path.isdir", MagicMock(return_value=True))
    @unittest.mock.patch("os.listdir")
    @unittest.mock.patch("os.stat")
    @unittest.mock.patch("apport.hookutils.read_file")
    def test_attach_dmi(
        self, read_file_mock: MagicMock, stat_mock: MagicMock, listdir_mock: MagicMock
    ) -> None:
        """attach_dmi()"""

        def stat(
            content: str = "private\n",
            st_mode: int = 0o100444,
            st_nlink: int = 1,
            st_size: int = 4096,
        ) -> tuple[str, os.stat_result]:
            st_time = 1698056204.0
            stat = os.stat_result(
                (st_mode, 1337, 22, st_nlink, 0, 0, st_size, st_time, st_time, st_time)
            )
            return (content, stat)

        dmi_files = {
            "bios_date": stat("07/20/2022\n"),
            "board_serial": stat(st_mode=0o100400),
            "uevent": stat("MODALIAS=dmi:[...]\n", st_mode=0o100644),
            "product_serial": stat(st_mode=0o100400),
            "product_name": stat("B550I AORUS PRO AX\n"),
            "sys_vendor": stat("Gigabyte Technology Co., Ltd.\n"),
            "power": stat(st_mode=0o40755, st_nlink=2, st_size=0),
            "bios_version": stat("F16e\n"),
            "bios_release": stat("5.17\n"),
            "board_vendor": stat("Gigabyte Technology Co., Ltd.\n"),
            "subsystem": stat(st_mode=0o40755, st_nlink=2, st_size=0),
            "product_family": stat("B550 MB\n"),
            "product_uuid": stat(
                "2cf4a3c2-8f76-4f38-95b5-77c8a9f6ec4f\n", st_mode=0o100400
            ),
            "bios_vendor": stat("American Megatrends International, LLC.\n"),
            "board_name": stat("B550I AORUS PRO AX\n"),
        }

        def mock_os_stat(path: str) -> os.stat_result:
            return dmi_files[os.path.basename(path)][1]

        def mock_read_file(path: str) -> str:
            return dmi_files[os.path.basename(path)][0].strip()

        listdir_mock.return_value = list(dmi_files.keys())
        stat_mock.side_effect = mock_os_stat
        read_file_mock.side_effect = mock_read_file
        report: dict[str, str] = {}
        apport.hookutils.attach_dmi(report)
        self.assertEqual(
            report,
            {
                "dmi.bios.date": "07/20/2022",
                "dmi.bios.release": "5.17",
                "dmi.bios.vendor": "American Megatrends International, LLC.",
                "dmi.bios.version": "F16e",
                "dmi.board.name": "B550I AORUS PRO AX",
                "dmi.board.vendor": "Gigabyte Technology Co., Ltd.",
                "dmi.product.family": "B550 MB",
                "dmi.product.name": "B550I AORUS PRO AX",
                "dmi.sys.vendor": "Gigabyte Technology Co., Ltd.",
                "MachineType": "Gigabyte Technology Co., Ltd. B550I AORUS PRO AX",
            },
        )

    @unittest.mock.patch("apport.hookutils.root_command_output")
    def test_attach_dmesg(self, root_command_output_mock):
        """attach_dmesg()"""
        root_command_output_mock.return_value = "[30804.972250] CPU0 is up"

        report = {}
        apport.hookutils.attach_dmesg(report)
        self.assertEqual(report, {"CurrentDmesg": "[30804.972250] CPU0 is up"})

        root_command_output_mock.assert_called_once_with(["dmesg"])

    def test_dmesg_overwrite(self):
        """attach_dmesg() does not overwrite already existing data"""
        report = {"CurrentDmesg": "existingcurrent"}

        apport.hookutils.attach_dmesg(report)
        self.assertEqual(report["CurrentDmesg"], "existingcurrent")

    @unittest.mock.patch("subprocess.run")
    @unittest.mock.patch("os.path.exists", MagicMock(return_value=True))
    def test_attach_journal_errors_with_date(self, run_mock):
        run_mock.return_value = subprocess.CompletedProcess(
            args=None, returncode=0, stdout=b"journalctl output", stderr=b""
        )
        now = datetime.datetime.now()

        report = apport.Report(date=now.strftime("%a %b %d %H:%M:%S %Y"))
        apport.hookutils.attach_journal_errors(report)

        self.assertEqual(run_mock.call_count, 1)
        self.assertEqual(report.get("JournalErrors"), "journalctl output")
        self.assertEqual(
            run_mock.call_args[0][0],
            [
                "journalctl",
                "--priority=warning",
                f"--since=@{int(now.timestamp()) - 10}",
                f"--until=@{int(now.timestamp()) + 10}",
            ],
        )

    @unittest.mock.patch("subprocess.run")
    @unittest.mock.patch("os.path.exists", MagicMock(return_value=True))
    def test_attach_journal_errors_without_date(self, run_mock):
        run_mock.return_value = subprocess.CompletedProcess(
            args=None, returncode=0, stdout=b"journalctl output", stderr=b""
        )

        report = apport.Report()
        del report["Date"]
        apport.hookutils.attach_journal_errors(report)

        self.assertEqual(run_mock.call_count, 1)
        self.assertEqual(report.get("JournalErrors"), "journalctl output")
        self.assertEqual(
            run_mock.call_args[0][0],
            ["journalctl", "--priority=warning", "-b", "--lines=1000"],
        )

    def test_path_to_key(self):
        """Transform a file path to a valid report key."""
        self.assertEqual(apport.hookutils.path_to_key("simple.txt"), "simple.txt")
        self.assertEqual(
            apport.hookutils.path_to_key("path/with/dirs.txt"), "path.with.dirs.txt"
        )
        self.assertEqual(
            apport.hookutils.path_to_key('/funny:characters!& ".txt'),
            ".funny.characters.._..txt",
        )

    @unittest.mock.patch("subprocess.Popen")
    @unittest.mock.patch("os.path.exists", MagicMock(return_value=True))
    def test_recent_syslog_journald_cmd(self, popen_mock):
        class _SkipPopen(Exception):
            pass

        popen_mock.side_effect = _SkipPopen

        cmd = ["journalctl", "--quiet", "-b", "-a"]
        cmd_system_only = cmd + ["--system"]

        with self.assertRaises(_SkipPopen):
            apport.hookutils.recent_syslog(re.compile("."))
        popen_mock.assert_called_once_with(cmd_system_only, stdout=unittest.mock.ANY)

        popen_mock.reset_mock()
        with self.assertRaises(_SkipPopen):
            apport.hookutils.recent_syslog(re.compile("."), journald_only_system=True)
        popen_mock.assert_called_once_with(cmd_system_only, stdout=unittest.mock.ANY)

        popen_mock.reset_mock()
        with self.assertRaises(_SkipPopen):
            apport.hookutils.recent_syslog(re.compile("."), journald_only_system=False)
        popen_mock.assert_called_once_with(cmd, stdout=unittest.mock.ANY)

    def test_deprecated_upstart_functions(self) -> None:
        """attach_upstart_*() throw deprecation warnings."""
        with self.assertWarns(PendingDeprecationWarning):
            apport.hookutils.attach_upstart_logs({}, "apport")
        with self.assertWarns(PendingDeprecationWarning):
            apport.hookutils.attach_upstart_overrides({}, "nonexisting")

    def test_recent_syslog_race_condition(self) -> None:
        """Test recent_syslog reads stdout buffer after process completion.

        LP: #2073935
        """
        with unittest.mock.patch("apport.hookutils.subprocess.Popen") as popen_mock:
            process = unittest.mock.Mock()
            popen_mock.return_value.__enter__.return_value = process
            process.returncode = 0

            process.poll = Mock(return_value=0)
            process.stdout = MagicMock()
            process.stdout.__iter__.side_effect = [iter([b"content"])]

            match_content = apport.hookutils.recent_syslog(
                re.compile("."), path="/nonexisting"
            )
            self.assertEqual(match_content, "content")

    def test_recent_syslog_long_process(self) -> None:
        """Test recent_syslog retrieves entire stdout."""
        with unittest.mock.patch("apport.hookutils.subprocess.Popen") as popen_mock:
            process = unittest.mock.Mock()
            popen_mock.return_value.__enter__.return_value = process
            process.returncode = 0

            process.poll = Mock(side_effect=[None, 0])
            process.stdout = MagicMock()
            process.stdout.__iter__.side_effect = [
                iter([b"test\n"]),
                iter([b"content\n"]),
            ]

            match_content = apport.hookutils.recent_syslog(
                re.compile("."), path="/nonexisting"
            )

            self.assertEqual(match_content, "test\ncontent\n")
